// f3dGrid.cpp
// Implementation of the f3dGrid<T> class for simple manipulation with scalar
// volumetric grids.
//
// This file is part of an f3d package, which was designed and partialy also
// implemented during the author's post doctoral stay at SUNY an Stony Brook,
// in the lab of Prof. Arie Kaufman.
// 
// For details, see the f3d World-Wide-Web page,
// http://www.viskom.oeaw.ac.at/~milos/page/Download.html
// or send a mail to milos.sramek@oeaw.ac.at
// 
// Copyright (C) 1999 Free Software Foundation
// Milos Sramek  <milos.sramek@oeaw.ac.at>
// 
// f3d package is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
// 
// the f3d package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
// See the GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public
// License along with the f3d package -- see the file COPYING.
// If not, write to the Free Software Foundation, Inc.,
// 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.


#include <f3dGridLite.h>
#include <assert.h>
#include <iostream>
#include <string>

using namespace::std;

extern int f3dErrorCode;
extern const char *f3dErrorMsg[];

// constructor: create the cubic grid
//template <class T, template <class T> class RASTER>
template <class T>
f3dGrid<T>::f3dGrid(int x, int y, int z, f3dDataType dt):
	nx(x), ny(y), nz(z)
{ 
	hdr = f3dDefaultHeader(x,y,z,dt,1);
	create();
	set(0); 
}


// constructor: create the grid
template <class T>
f3dGrid<T>::f3dGrid(f3dHeader &h):
	nx(h.nx), ny(h.ny), nz(h.nz)
{ 
	hdr = f3dDupHeader(&h);
	create();
}


// constructor: load the grid from a file
	template <class T>
f3dGrid<T>::f3dGrid(char *name, f3dDataType dt)
{
	FILE *fp;
	fp = fopen(name,"rb");
	if(!fp){
		fprintf(stderr,"f3dGrid<T>::f3dGrid : %s -- File open error\n",
				name);
		exit(1);
	}

	fp = f3dReadHeaderF(fp, &hdr);
	if(fp == NULL){
		cerr<<"ERROR -- f3dGrid<T>::f3dGrid -- "<<
			f3dErrorMsg[f3dErrorCode]<<endl;
		exit(1);
	}
	nx = hdr.nx;
	ny = hdr.ny;
	nz = hdr.nz;
	create();

	loadFromFile(fp, dt);
	fclose(fp);
}

// constructor: load the grid from a file
	template <class T>
f3dGrid<T>::f3dGrid(FILE *fp, f3dDataType dt)
{
	fp = f3dReadHeaderF(fp, &hdr);
	nx = hdr.nx;
	ny = hdr.ny;
	nz = hdr.nz;
	create();
	loadFromFile(fp, dt);
}

// constructor: load the grid from an open file
	template <class T>
void f3dGrid<T>::loadFromFile(FILE *fp, f3dDataType dt) 
{
	if(dt > f3dDoubleType ){
		cerr<<"f3dGrid::loadFromFile -- invalid voxel type ("<<dt<<")\n";
		exit(1);
	}

	if(hdr.dtype != dt){
		cerr<<"f3dGrid<T>::loadFromFile -- Voxel type conflict"<<endl;
		exit(1);  
	}

	if(sizeof(T) != f3dVoxelSize(&hdr)){
		cerr<<"f3dGrid<T>::loadFromFile -- Voxel size mismatch"<<endl;
		exit(1);  
	}

	// load 
	// reversed band-slice order since f3dformat6
	for(int z=0; z<hdr.nz; z++){
		f3dReadSlice(fp,data+z*nx*ny, hdr.endian, f3dVoxelSize(&hdr), hdr.nx*hdr.ny, hdr.comptype);
	}
}

// Creates a new grid. Assumes correct initialization of class parameters
	template <class T>
void f3dGrid<T>::create(void)
{
	data = new T[nx*ny*nz];
	if(!data){
		fprintf(stderr,"f3dGrid<T>::create -- Allocation error\n");
		exit(1);
	}
}

// Destroys the grid
	template <class T>
void f3dGrid<T>::destroy(void) 
{
	delete [] data;
	f3dDelHeader(&hdr);
}

// copy function for the Data
	template <class T>
void f3dGrid<T>::copy(f3dGrid<T> const& other)
{
	nx = other.nx;
	ny = other.ny;
	nz = other.nz;
	create();
	memcpy(other.data, data, nx*ny*nz*sizeof(T));
	hdr = f3dDupHeader(&other.hdr);
}

/* 
 *  write 3D volume of data in f3d format
 */
template <class T>
void f3dGrid<T>::save(const char *name) const
{
	FILE *fptr;

	fptr = fopen(name,"wb");
	if(!fptr){
		fprintf(stderr,"f3dGrid<t>::save : %s -- File open error\n",
				name);
		exit(1);
	}
	save(fptr);
	fclose(fptr);
}

/* 
 *  write 3D volume of data in f3d format
 */
template <class T>
void f3dGrid<T>::save(FILE *fp) const
{
	int i,b,z;

	f3dHeader h(f3dDupHeader(&hdr));

	// the icon image
	unsigned char *cbuf = new unsigned char[nx*ny]; 
	assert(cbuf);

	f3dWriteHeader(fp, &h, cbuf, nx, ny);
	delete [] cbuf;

	// write the slices now
	// reversed band-slice order since f3dformat6
	for(z=0; z<nz; z++){
		f3dWriteSlice(fp,data+nx*ny*z, f3dVoxelSize(&h), nx*ny, h.comptype);
	}
}


// Explicit instantiation of template <class T>

template class f3dGrid<f3dUChar>;
template class f3dGrid<f3dUInt16>;
template class f3dGrid<f3dUInt32>;
template class f3dGrid<f3dFloat>;
template class f3dGrid<f3dDouble>;
template class f3dGrid<f3dInt32>;
template class f3dGrid<f3dInt16>;
template class f3dGrid<f3dChar>;
