// Definition of the f3dGrid<T> class for simple manipulation with 
// volumetric grids.
//
// This file is part of an f3d package, which was designed and partialy also
// implemented during the author's post doctoral stay at SUNY an Stony Brook,
// in the lab of Prof. Arie Kaufman.
// 
// For details, see the f3d World-Wide-Web page,
// http://www.viskom.oeaw.ac.at/~milos/page/Download.html
// or send a mail to milos.sramek@oeaw.ac.at
// 
// Copyright (C) 1999 Free Software Foundation
// Milos Sramek  <milos.sramek@oeaw.ac.at>
// 
// f3d package is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
// 
// the f3d package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
// See the GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public
// License along with the f3d package -- see the file COPYING.
// If not, write to the Free Software Foundation, Inc.,
// 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifndef _F3DGRID_LITE_H_
#define _F3DGRID_LITE_H_

#include"f3d.h"

/** A class for manipulation with 3D arrays of multiband voxels.
The basic grid is padded by {\tt bl} thick layer of voxels, in order to enable
filtering with large filter kernels.
Although, theoreticaly, voxels can be of any type, due to explicit 
instantiation (in file f3dGrid.cpp) the following types are allowed:
\begin{itemize}
\item {\tt f3dGrid<f3dFloat>}
\item {\tt f3dGrid<f3dInt32>}
\item {\tt f3dGrid<f3dUInt32>}
\item {\tt f3dGrid<f3dInt16>}
\item {\tt f3dGrid<f3dIntHU>}
\item {\tt f3dGrid<f3dUInt16>}
\item {\tt f3dGrid<f3dUInt12>}
\item {\tt f3dGrid<f3dChar>}
\item {\tt f3dGrid<f3dUChar>}
\end{itemize}
@author Milos Sramek, SUNY Stony Brook, Center for Visual Computing
*/
//template <class T, template <class T> class RASTER> class f3dGrid: public f3dVolume 
template <class T> class f3dGrid
{
private:
	f3dHeader hdr;     		// f3d header describing the grid
	T *data;			// the data
	int nx, ny, nz;
public:

	/**@name  Constructors, assignment and destructor*/
	//@{
	/** Constructor: define an x {\tt x} y {\tt x} z grid, 
	  cubic by default, padded by l voxels thick layer.\\
	@param nx dimension in the column direction
	@param ny dimension in the row direction
	@param nz dimension in the slice direction
	@param dtype basic voxel type (f3dUCharType, ....) defined in f3d.h. Must correspond to class template type.
	*/
	f3dGrid(int nx, int ny, int nz, f3dDataType dtype);

	/** Constructor: define a grid by means of a header.
	@param h the header
	*/
	f3dGrid(f3dHeader &h);

	/** Constructor: load the grid from a file.\\
	@param name file name.
	@param dtype basic voxel type (f3dUCharType, ...) defined in {\tt f3d.h}. Must correspond to class template type and the file type;
	*/
	f3dGrid(char *name, f3dDataType dtype);

	/** Constructor: load the grid from an open file.\\
	@param fp open file
	@param dtype basic voxel type (f3dUCharType, ...) defined in {\tt f3d.h}. Must correspond to class template type and the file type;
	*/
	f3dGrid(FILE *fp, f3dDataType dtype);

	/// copy constructor
	f3dGrid(f3dGrid<T> const &other) 
	{ copy(other); }

	/// overloaded assignment
	f3dGrid const& operator=(f3dGrid<T> const &other)
	{
		if(this != &other){ 
			destroy();
			copy(other); 
		}
		return *this;
	}

	/// Duplicate the grid
	f3dGrid * Dup() const { return new f3dGrid<T>(*this); }

	/// destructor 
	~f3dGrid(){destroy();}
//@}

/**@name Messages for reading and seting  grid matrix parameters */
//@{
	/// X resolution in pixels
	int getNX(int band=0) const {return hdr.nx;}

	/// Y resolution in pixels
	int getNY(int band=0) const {return hdr.ny;}

	/// Z resolution in pixels
	int getNZ(int band=0) const {return hdr.nz;}
//@}

/**@name Data access methods */
//@{

	/// get voxel value by coordinates
	T get(int x, int y, int z) const { return *(data + getShift(x, y, z)); }

	/** get voxel value by coordinates. 
	 * Mirrored value is returned for points outside the volume
 	*/
	T getWrap(int x, int y, int z) const { 
		if(x<0) x = -x; 
		if(y<0) y = -y; 
		if(z<0) z = -z;
		if(x>=getNX()) x = 2*getNX() - x - 2;
		if(y>=getNY()) y = 2*getNY() - y - 2;
		if(z>=getNZ()) z = 2*getNZ() - z - 2;
		return get(x, y, z); 
	}

	/// set voxel value by coordinates
	void set(int x, int y, int z, const T val) 
	{ *(data + getShift(x, y, z)) = val; }

	/// set the volume to one value
	void set(T const val) {for(int i=0; i<nx*ny*nz; i++) *(data+i) = val;}
//@}

/**@name {\large Save/load  grid to/from file} */
//@{

	/// save the volume into file in f3d format
	void save(const char *name) const;

	/// save the volume into file in f3d format
	void save(FILE *fp) const;
//@}

/**@name Access to header comment lines */
//@{

	/// add header comment
	void addComm(const char *txt) { f3dAddHdrComment(&hdr, txt);}

	/// set header comments
	void setComm(int n, char **txt) { f3dSetHdrComment(&hdr, n, txt);}

	/// get number of comment lines
	int getNComm(void) const {return hdr.commentLines;}

	/// get pointer to n-th comment line
	char *getComm(int n) const {return hdr.comment[n];}
//@}

/**@name Other methods */
//@{
	/// get the whole header
	f3dHeader getHeader(void) const {return hdr;}
//@}

private:
/// build the class from data taken from an open file
void loadFromFile(FILE *fp, f3dDataType dt);

/// get displacement (shift) of a voxel with respect to the data origin
int getShift(int x, int y, int z) const { return z*nx*ny + y*nx + x;}
///  copy the grid
void copy(f3dGrid const &other);

/// create the grid
void create(void); 

/// destroy the grid
void destroy(void);
};

#endif //_F3DGRID_LITE_H_
